from minio_manager.classes.errors import MinioInvalidIamCredentialsError, MinioMalformedIamPolicyError
from minio_manager.classes.mc_wrapper import McWrapper
from minio_manager.classes.minio_resources import ServiceAccount
from minio_manager.clients import get_controller_user_policy, get_mc_wrapper, get_minio_config, get_secret_manager
from minio_manager.utilities import compare_objects, logger


def service_account_exists(client: McWrapper, account: ServiceAccount):
    try:
        if account.access_key:
            client.service_account_info(account.access_key)
            return True
    except MinioInvalidIamCredentialsError as e:
        # account does not exist in MinIO
        logger.debug(f"Error for {account.access_key}: {e}")

    logger.debug(f"Access key for {account.name} not found in secret backend, trying to find it in MinIO.")
    sa_list = client.service_account_list(client.cluster_controller_user)
    for sa in sa_list:
        sa_info = client.service_account_info(sa["accessKey"])
        if hasattr(sa_info, "name") and sa_info["name"] == account.name:
            logger.debug(f"Found access key '{sa_info['accessKey']}' for '{account.name}' in MinIO: {sa_info}")
            return True

    return False


def apply_base_policy(account: ServiceAccount):
    client = get_mc_wrapper()
    account.generate_service_account_policy()
    client.service_account_set_policy(account.access_key, str(account.policy_file))


def handle_sa_policy(account: ServiceAccount):
    """
    Manage policies for service accounts.

    Compare the desired policy with what is currently applied, and update if needed.
    Retrieves the updated policy and compares it again. If it does not match, compare to the controller user's policy.
    If those match, the supplied policy had more permissions than allowed, and we have to revert to the base policy.

    Args:
        account (ServiceAccount)
    """
    client = get_mc_wrapper()
    desired_policy = account.policy

    current_policy = client.service_account_get_policy(account.access_key)

    policies_diff_pre = compare_objects(current_policy, desired_policy)
    if not policies_diff_pre:
        return

    logger.debug(f"Updating service account policy for '{account.name}'.")
    try:
        client.service_account_set_policy(account.access_key, str(account.policy_file))
    except MinioMalformedIamPolicyError:
        logger.error(
            f"Policy for service account '{account.name}' is malformed, reverting to base policy for service account."
        )
        apply_base_policy(account)
        return

    # TODO: so this works as intended, but MinIO somehow now seems to accept a policy with more permissions?
    #  it won't actually allow the actions, but the policy is still valid and now won't get updated..?
    current_updated_policy = client.service_account_get_policy(account.access_key)
    policies_diff_post = compare_objects(current_updated_policy, desired_policy)
    if not policies_diff_post:
        return

    logger.warning(f"Applying policy for service account '{account.name}' failed.")
    logger.debug("Retrieving controller user credentials...")
    controller_user_policy = get_controller_user_policy()
    logger.debug("Comparing controller user policy to currently applied policy...")
    policies_diff_fallback = compare_objects(controller_user_policy, current_updated_policy)
    if policies_diff_fallback:
        logger.critical("Unknown situation where the live service account policy")
        logger.critical("a) does not match what we tried to apply;")
        logger.critical("b) also does not match to the controller user's policy, which it should fall back to if")
        logger.critical("the policy we tried to apply has more permissions than the controller user's.")

    logger.warning(f"Reverting to base policy for service account '{account.name}'")
    apply_base_policy(account)


def handle_service_account(account: ServiceAccount):
    """
    Manage service accounts.

    Steps taken:
    1) check if service account exists in minio
    2) check if service account exists in secret backend
    3) if it exists in MinIO but not the secret backend, throw an error and return
    4) if it exists in the secret backend but not in MinIO, create it using the secret backend credentials
    5) if it does not, create service account in minio and secret backend
    6) if a policy_file is configured, load it and try to apply it for the service account

    Args:
        account (ServiceAccount)
    """
    client = get_mc_wrapper()
    secrets = get_secret_manager(get_minio_config())

    # Determine if access key credentials exists in secret backend
    credentials = secrets.get_credentials(account.name)
    # Determine if access key exists in MinIO
    sa_exists = service_account_exists(client, credentials)

    # Scenario 1: service account exists in MinIO but not in secret backend
    if sa_exists and not credentials.access_key:
        logger.error(
            f"Service account {account.name} exists in MinIO but not in secret backend! Manual intervention required."
        )
        logger.error(
            "Either find the credentials elsewhere and add them to the secret backend, or delete the service "
            "account from MinIO and try again."
        )
        return

    # Scenario 2: service account exists in secret backend but not in MinIO
    if credentials.secret_key and not sa_exists:
        logger.warning(
            f"Service account {account.name} exists in secret backend but not in MinIO. Using existing credentials."
        )
        client.service_account_add(credentials)
        sa_exists = True
        logger.info(f"Created service account '{credentials.name}', access key: {credentials.access_key}")

    # Scenario 3: service account does not exist in neither MinIO nor the secret backend
    if not sa_exists and not credentials.access_key:
        # TODO: catch scenario where an access key is deleted in MinIO, but MinIO does not accept the creation of a
        #  service account with the same access key, which sometimes happens.
        # Create the service account in MinIO
        credentials = client.service_account_add(credentials)
        # Create credentials in the secret backend
        secrets.set_password(credentials)
        logger.info(f"Created service account '{credentials.name}' with access key '{credentials.access_key}'")

    account.access_key = credentials.access_key
    account.secret_key = credentials.secret_key

    if account.policy_file:
        handle_sa_policy(account)



from __future__ import annotations

import json
import sys
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import ClassVar

from minio.lifecycleconfig import LifecycleConfig
from minio.versioningconfig import VersioningConfig

from minio_manager.utilities import get_env_var, logger, module_directory, read_json

sa_policy_embedded = f"{module_directory}/resources/service-account-policy-base.json"
sa_policy_base_file = get_env_var("MINIO_MANAGER_SERVICE_ACCOUNT_POLICY_BASE_FILE", sa_policy_embedded)


class Bucket:
    def __init__(
        self,
        name: str,
        create_service_account: bool = True,
        versioning: VersioningConfig | None = None,
        lifecycle_config: LifecycleConfig | None = None,
    ):
        self.name = name
        self.create_sa = create_service_account
        self.versioning = versioning
        self.lifecycle_config = lifecycle_config


class BucketPolicy:
    # TODO: try loading the policy file in order to validate its contents
    def __init__(self, bucket: str, policy_file: str):
        self.bucket = bucket
        self.policy_file = policy_file


class ServiceAccount:
    policy = ClassVar[dict]

    def __init__(
        self,
        name: str,
        description: str = "",
        access_key: str | None = None,
        secret_key: str | None = None,
        policy: dict | None = None,
        policy_file: Path | str | None = None,
    ):
        if len(name) > 32:
            self.name = name[:32]
            self.description = name + " " + description
        else:
            self.name = name
            self.description = description
        self.access_key = access_key
        self.secret_key = secret_key
        if policy_file:
            if isinstance(policy_file, Path):
                self.policy_file = policy_file
            else:
                self.policy_file = Path(policy_file)
        else:
            self.policy_file = None
        if policy:
            self.policy = policy
        elif self.policy_file:
            try:
                self.policy = read_json(self.policy_file)
            except FileNotFoundError:
                logger.critical(f"Policy file '{self.policy_file}' for service account '{name}' not found!")
                sys.exit(1)

    def generate_service_account_policy(self):
        """
        Generate a policy for a service account that gives access to a bucket with the same name as the service account.
        """
        with Path(sa_policy_base_file).open() as base:
            base_policy = base.read()

        temp_file = NamedTemporaryFile(prefix=self.name, suffix=".json", delete=False)
        with temp_file as out:
            new_content = base_policy.replace("BUCKET_NAME_REPLACE_ME", self.name)
            out.write(new_content.encode("utf-8"))

        self.policy = json.loads(new_content)
        self.policy_file = Path(temp_file.name)


class IamPolicy:
    def __init__(self, name: str, policy_file: str):
        self.name = name
        self.policy_file = policy_file


class IamPolicyAttachment:
    def __init__(self, username: str, policies: list):
        self.username = username
        self.policies = policies



import logging
import re
from logging import Filter, Formatter, LogRecord

RED = "\033[0;31m"
GREEN = "\033[0;32m"
YELLOW = "\033[0;33m"
BLUE = "\033[0;34m"
PURPLE = "\033[0;35m"
BOLD = "\033[1m"
RESET = "\033[0m"

COLORS = {
    "DEBUG": PURPLE,
    "INFO": GREEN,
    "WARNING": YELLOW,
    "ERROR": RED,
    "CRITICAL": BOLD + RED,
}


class MinioManagerFilter(Filter):
    wrapper_secret_re = re.compile(r"--secret-key (?P<secret>[\w+/]*)")
    alias_set_secret_re = re.compile(r"alias set .+ (?P<secret>[\w+/]*)$")

    def filter(self, record: LogRecord) -> bool:
        if "--secret" in record.msg:
            record.msg = self.mask_secret(record.msg, self.wrapper_secret_re)
        if "alias set" in record.msg:
            record.msg = self.mask_secret(record.msg, self.alias_set_secret_re)

        return True

    @staticmethod
    def mask_secret(message: str, regex: re.Pattern) -> str:
        result = regex.search(message)
        if result:
            message = message.replace(result.group("secret"), "************")
        return message


class MinioManagerFormatter(Formatter):
    def __init__(self, log_level: int):
        self.log_level = log_level
        if log_level is logging.INFO:
            log_format = "[{asctime}] [{levelname:^8s}] {message}"
            super().__init__(fmt=log_format, datefmt="%Y-%m-%d %H:%M:%S", style="{")
        else:
            log_format = "[{asctime}] [{levelname:^8s}] [{filename:>26s}:{lineno:<4d} - {funcName:<24s} ] {message}"
            super().__init__(fmt=log_format, style="{")

    def format(self, record: LogRecord):  # noqa: A003
        if record.levelname in COLORS:
            record.msg = COLORS[record.levelname] + record.msg + RESET
            # record.levelname = COLORS[record.levelname] + record.levelname + RESET

        return super().format(record)


class MinioManagerLogger(logging.Logger):
    def __init__(self, name: str, log_level: str):
        super().__init__(name)
        if log_level == "INFO":
            self.setLevel(logging.INFO)
        else:
            self.setLevel(logging.DEBUG)

        handler = logging.StreamHandler()
        formatter = MinioManagerFormatter(self.level)
        this_filter = MinioManagerFilter()
        handler.setFormatter(formatter)
        handler.addFilter(this_filter)
        self.addHandler(handler)



from __future__ import annotations

import json
import sys
from pathlib import Path

from minio.commonconfig import Filter
from minio.lifecycleconfig import Expiration, LifecycleConfig, NoncurrentVersionExpiration, Rule
from minio.versioningconfig import VersioningConfig as VeCo

from minio_manager.classes.minio_resources import Bucket, BucketPolicy, IamPolicy, IamPolicyAttachment, ServiceAccount
from minio_manager.utilities import get_env_var, logger, read_yaml

default_bucket_versioning = get_env_var("MINIO_MANAGER_DEFAULT_BUCKET_VERSIONING", "Suspended")
default_bucket_lifecycle_policy = get_env_var("MINIO_MANAGER_DEFAULT_LIFECYCLE_POLICY", "")
default_bucket_create_service_account = get_env_var("MINIO_MANAGER_DEFAULT_BUCKET_CREATE_SERVICE_ACCOUNT", "True")
default_bucket_allowed_prefix = get_env_var("MINIO_MANAGER_ALLOWED_BUCKET_PREFIX", "")


class ClusterResources:
    """ClusterResources is the object containing all the cluster resources:

    - buckets
    - bucket_policies
    - service_accounts
    - iam_policies
    - iam_policy_attachments
    """

    buckets: list[Bucket]
    bucket_policies: list[BucketPolicy]
    service_accounts: list[ServiceAccount]
    iam_policies: list[IamPolicy]
    iam_policy_attachments: list[IamPolicyAttachment]

    def parse_buckets(self, buckets: list[dict]) -> list[Bucket]:
        """Parse the provided buckets with the following steps:
        For each provided bucket
            1. check the provided versioning. If versioning is not provided, set the default.
            2. check if an object lifecycle JSON file is provided, use the default_bucket_lifecycle_policy, or skip OLM
            3. parse the file and create a LifecycleConfig object for the bucket
            4. create a Bucket object

        Args:
            buckets: list of buckets to parse

        Returns:
            [Bucket]: list of Bucket objects
        """
        if not buckets:
            logger.info("No buckets configured, skipping.")
            return []

        bucket_objects = []
        if not isinstance(buckets, list):
            logger.error("buckets must be defined as a list!")
            sys.exit(1)

        lifecycle_config = self.parse_bucket_lifecycle_file(default_bucket_lifecycle_policy)
        bucket_names = []

        for bucket in buckets:
            name = bucket["name"]
            if name in bucket_names:
                logger.error(f"Bucket '{name}' defined multiple times. Stopping.")
                sys.exit(1)
            bucket_names.append(name)
            logger.debug(f"Parsing bucket {name}")
            if not name.startswith(default_bucket_allowed_prefix):
                logger.error(f"Bucket {name} does not start with required prefix '{default_bucket_allowed_prefix}'.")
                sys.exit(1)

            versioning = bucket.get("versioning")
            try:
                versioning_config = VeCo(versioning) if versioning else VeCo(default_bucket_versioning)
            except ValueError as ve:
                logger.error(f"Error parsing versioning setting: {' '.join(ve.args)}")
                sys.exit(1)
            create_sa = bool(bucket.get("create_service_account", default_bucket_create_service_account))
            lifecycle_file = bucket.get("object_lifecycle_file")
            if lifecycle_file:
                bucket_lifecycle = self.parse_bucket_lifecycle_file(lifecycle_file)
                if isinstance(bucket_lifecycle, LifecycleConfig):
                    lifecycle_config = bucket_lifecycle
            bucket_objects.append(Bucket(name, create_sa, versioning_config, lifecycle_config))

        return bucket_objects

    def parse_bucket_lifecycle_file(self, lifecycle_file: str) -> LifecycleConfig | None:
        """Parse a bucket lifecycle config file.
        The config files must be in JSON format and can be best obtained by running the following command:
            mc ilm rule export $cluster/$bucket > $policy_file.json

        Args:
            lifecycle_file: lifecycle config file

        Returns:
            LifecycleConfig object
        """
        if not lifecycle_file:
            return

        rules: list = []

        try:
            with Path(lifecycle_file).open() as f:
                config_data = json.load(f)
        except FileNotFoundError:
            logger.error(f"Lifecycle file {lifecycle_file} not found, skipping configuration.")
            sys.exit(1)
        except PermissionError:
            logger.error(f"Incorrect file permissions on {lifecycle_file}, skipping configuration.")
            sys.exit(1)

        try:
            rules_dict = config_data["Rules"]
        except KeyError:
            logger.error(f"Lifecycle file {lifecycle_file} is missing the required 'Rules' key.")
            sys.exit(1)
        try:
            for rule_data in rules_dict:
                parsed_rule = self.parse_bucket_lifecycle_rule(rule_data)
                rules.append(parsed_rule)
        except AttributeError:
            logger.error(f"Error parsing lifecycle file {lifecycle_file}. Is the format correct?")
            sys.exit(1)

        if not rules:
            return

        return LifecycleConfig(rules)

    @staticmethod
    def parse_bucket_lifecycle_rule(rule_data: dict) -> Rule:
        """Parse a single bucket object lifecycle rule
        TODO: implement date and days in Expiration, implement Transition, NoncurrentVersionTransition, Filter, and
          AbortIncompleteMultipartUpload

        Args:
            rule_data: dict with rule data

        Returns:
            Rule
        """
        rule_dict = {"status": rule_data.get("Status"), "rule_id": rule_data.get("ID")}

        expiration = rule_data.get("Expiration")
        if expiration:
            expire_delete_marker = expiration.get("ExpiredObjectDeleteMarker")
            rule_dict["expiration"] = Expiration(expired_object_delete_marker=expire_delete_marker)

        noncurrent_version_expiration = rule_data.get("NoncurrentVersionExpiration")
        if noncurrent_version_expiration:
            noncurrent_expire_days = noncurrent_version_expiration.get("NoncurrentDays")
            rule_dict["noncurrent_version_expiration"] = NoncurrentVersionExpiration(noncurrent_expire_days)

        # An empty filter is required for the rule to be valid
        rule_dict["rule_filter"] = Filter(prefix="")

        rule = Rule(**rule_dict)
        return rule

    @staticmethod
    def parse_bucket_policies(bucket_policies):
        if not bucket_policies:
            logger.info("No bucket policies configured, skipping.")
            return []

        bucket_policy_objects = []
        for bucket_policy in bucket_policies:
            bucket_policy_objects.append(BucketPolicy(bucket_policy["bucket"], bucket_policy["policy_file"]))

        return bucket_policy_objects

    @staticmethod
    def parse_service_accounts(service_accounts):
        if not service_accounts:
            logger.info("No service accounts configured, skipping.")
            return []

        service_account_objects, service_account_names = [], []
        for service_account in service_accounts:
            name = service_account["name"]
            if name in service_account_names:
                logger.error(f"Service account '{name}' defined multiple times. Stopping.")
                sys.exit(1)
            service_account_names.append(name)
            policy_file = service_account.get("policy_file")
            sa_obj = ServiceAccount(name=name, policy_file=policy_file)
            service_account_objects.append(sa_obj)

        return service_account_objects

    @staticmethod
    def parse_iam_attachments(iam_policy_attachments):
        if not iam_policy_attachments:
            logger.info("No IAM policy attachments configured, skipping.")
            return []

        iam_policy_attachment_objects = []
        for user in iam_policy_attachments:
            iam_policy_attachments.append(IamPolicyAttachment(user["username"], user["policies"]))

        return iam_policy_attachment_objects

    @staticmethod
    def parse_iam_policies(iam_policies):
        if not iam_policies:
            logger.info("No IAM policies configured, skipping.")
            return []

        iam_policy_objects, iam_policy_names = [], []
        for iam_policy in iam_policies:
            name = iam_policy["name"]
            if name in iam_policy_names:
                logger.error(f"IAM policy '{name}' defined multiple times. Stopping.")
                sys.exit(1)
            iam_policy_names.append(name)
            iam_policy_objects.append(IamPolicy(name, iam_policy["policy_file"]))

        return iam_policy_objects

    def parse_resources(self, resources_file: str):
        try:
            resources = read_yaml(resources_file)
        except FileNotFoundError:
            logger.error(f"Resources file {resources_file} not found. Stopping.")
            sys.exit(1)
        except PermissionError:
            logger.error(f"Incorrect file permissions on {resources_file}. Stopping.")
            sys.exit(1)

        buckets = resources.get("buckets")
        self.buckets = self.parse_buckets(buckets)

        bucket_policies = resources.get("bucket_policies")
        self.bucket_policies = self.parse_bucket_policies(bucket_policies)

        service_accounts = resources.get("service_accounts")
        self.service_accounts = self.parse_service_accounts(service_accounts)

        iam_policies = resources.get("iam_policies")
        self.iam_policies = self.parse_iam_policies(iam_policies)

        iam_policy_attachments = resources.get("iam_policy_attachments")
        self.iam_policy_attachments = self.parse_iam_attachments(iam_policy_attachments)


class MinioConfig:
    """MinioConfig is the MinIO server configuration object containing the connection details."""

    def __init__(self):
        self.name = get_env_var("MINIO_MANAGER_CLUSTER_NAME")
        self.endpoint = get_env_var("MINIO_MANAGER_S3_ENDPOINT")
        self.secure = get_env_var("MINIO_MANAGER_S3_ENDPOINT_SECURE", True)
        self.controller_user = get_env_var("MINIO_MANAGER_MINIO_CONTROLLER_USER")
        self.access_key = None
        self.secret_key = None
        self.cluster_resources = get_env_var("MINIO_MANAGER_CLUSTER_RESOURCES_FILE", "resources.yaml")
        self.secret_backend_type = get_env_var("MINIO_MANAGER_SECRET_BACKEND_TYPE")
        self.secret_s3_bucket = get_env_var("MINIO_MANAGER_SECRET_BACKEND_S3_BUCKET", "minio-manager-secrets")



import json
import logging
import os
from pathlib import Path

import yaml
from deepdiff import DeepDiff

from minio_manager.classes.logging_config import MinioManagerLogger

logger = logging.getLogger("root")
logger_setup = False  # whether the logger is already configured or not
module_directory = os.path.dirname(__file__)


def read_yaml(file: str | Path) -> dict:
    with open(file) as f:
        return yaml.safe_load(f)


def read_json(file) -> dict:
    with open(file) as f:
        return json.load(f)


def setup_logging():
    global logger
    log_level = get_env_var("MINIO_MANAGER_LOG_LEVEL", "INFO")
    log_name = "root" if log_level == "DEBUG" else "minio-manager"
    logger = MinioManagerLogger(log_name, log_level)
    logger.debug(f"Configured log level: {log_level}")


def compare_objects(a: dict, b: dict, ignore_order: bool = True) -> bool | dict:
    """Compare two dicts and return False if they match, the differences if they don't"""
    result = DeepDiff(a, b, ignore_order=ignore_order)
    if result:
        return result
    return False


def get_env_var(name: str, default=None) -> str:
    """
    Get an environment variable and strip any leading and trailing single and double quotes.
    This is because Python literally loads them.

    Args:
        name: str, the name of the environment variable
        default: str, default if not set

    Returns: str stripped
    """
    try:
        variable = os.environ[name]
        strip_double_quotes = variable.strip('"')
        return strip_double_quotes.strip("'")
    except KeyError:
        if default is None:
            logger.critical(f"Required environment variable {name} not found!")
            exit(1)

    return default


if not logger_setup:
    setup_logging()
    logger_setup = True
